/* 
   CS 202, Fall 2007
   Homework#7, Problem 1

   Author: Harish Kumar
*/
/* Modifying the load() function of the Simpletron simulator (HW#3, 
	Problem 3) to read and load SML programs from a text file 
	named "prog.sml".
*/

#include <iostream>
#include <iomanip>
using namespace std; // Indication of Laziness

#include <fstream> // file stream
using std::ifstream; // input file stream

#include <cstdlib>
using std::exit; // exit function prototype

const int SIZE = 100, MAX_WORD = 9999, MIN_WORD = -9999;
const long SENTINEL = -99999;
enum Commands { READ = 10, WRITE, LOAD = 20, STORE, ADD = 30, SUBTRACT,
                DIVIDE, MULTIPLY, BRANCH = 40, BRANCHNEG, BRANCHZERO, HALT };

void load( int * const );
void execute( int * const, int * const, int * const, int * const, 
              int * const, int * const);
void dump( const int * const, int, int, int, int, int );
bool validWord( int );

int main()
{
   int memory[ SIZE ] = { 0 }, accumulator = 0, instructionCounter = 0, 
       opCode = 0, operand = 0, instructionRegister = 0;

   load( memory );
   
   execute( memory, &accumulator, &instructionCounter, &instructionRegister,
           &opCode, &operand );
   dump( memory, accumulator, instructionCounter, instructionRegister,
        opCode, operand );
   
   return 0;
}

// Modified load function
void load( int * const loadMemory )
{
    long instruction;
    int i = 0; // used to keep track of line numbers
	char comment[61]; // to store comments
    
	// ifstream constructor opens the file
	ifstream inSmlFile( "prog.sml", ios::in );

	// exit program if ifstream could not open file
	if( !inSmlFile )
	{
		cerr << "File could not be opened" << endl;
		exit( 1 );
	} // end if
   
	cout << "***           Harish Welcome to Simpletron          ***\n" // displaying my name
         << "***            Loading File prog.sml                ***\n"; // displaying program 
name rather static

   inSmlFile >> instruction; // reading operation code into instruction variable
   inSmlFile.getline( comment, 61 ); // loading comments after opcode into character array 
comment
   cout << setw( 2 ) << setfill( '0' ) << i << ") " << instruction << comment << endl; // 
printing out the first line from the file
   
   while ( instruction != SENTINEL ) {
      
      if ( !validWord( instruction ) )
         cout << "Number out of range. Please enter again.\n";
      else
         loadMemory[ i++ ] = instruction;

      
	  inSmlFile >> instruction; // reading the opcode into instruction
	  inSmlFile.getline( comment, 61 ); // reading the comments into char array comment
	  // function setfill sets the padding character for unused
      // field widths. 
      cout << setw( 2 ) << setfill( '0' ) << i << ") " << instruction << comment << endl; // 
printing out the next line from the file
      
   }
}

void execute( int * const memory, int * const acPtr, int * const icPtr, 
              int * const irPtr, int * const opCodePtr, int * const opPtr )
{
   bool fatal = false;
   int temp;
   const char *messages[] = { "Accumulator overflow          ***",
                              "Attempt to divide by zero     ***",
                              "Invalid opcode detected       ***" },
         *termString = "\n*** Simpletron execution abnormally terminated ***",
         *fatalString = "*** FATAL ERROR: ";

   cout << "\n************START SIMPLETRON EXECUTION************\n\n";

   do {
      *irPtr = memory[ *icPtr ];
      *opCodePtr = *irPtr / 100;
      *opPtr = *irPtr % 100;
      
      switch ( *opCodePtr ) {
         case READ:
            cout << "Enter an integer: ";
            cin >> temp;
            
            while ( !validWord( temp ) ) {
               cout << "Number out of range. Please enter again: ";
               cin >> temp;
            }
            
            memory[ *opPtr ] = temp;
            ++( *icPtr );
            break;
         case WRITE:
            cout << "Contents of " << setw( 2 ) << setfill( '0' ) << *opPtr
                 << ": " << memory[ *opPtr ] << '\n';
            ++( *icPtr );
            break;
         case LOAD:
            *acPtr = memory[ *opPtr ];
            ++( *icPtr );
            break;
         case STORE:
            memory[ *opPtr ] = *acPtr;
            ++( *icPtr );
            break;
         case ADD:
            temp = *acPtr + memory[ *opPtr ];
            
            if ( !validWord( temp ) ) {
               cout << fatalString << messages[ 0 ] << termString << '\n';
               fatal = true;
            }
            else {
               *acPtr = temp;
               ++( *icPtr );
            }
               
            break;
         case SUBTRACT:
            temp = *acPtr - memory[ *opPtr ];
            
            if ( !validWord( temp ) ) {
               cout << fatalString << messages[ 0 ] << termString << '\n';
               fatal = true;
            }
            else {
               *acPtr = temp;
               ++( *icPtr );
            }
               
            break;
         case DIVIDE:
            if ( memory[ *opPtr ] == 0 ) {
               cout << fatalString << messages[ 1 ] << termString << '\n';
               fatal = true;
            }
            else {
               *acPtr /= memory[ *opPtr ];
               ++( *icPtr );
            }
            
            break;
         case MULTIPLY:
            temp = *acPtr * memory[ *opPtr ];
            
            if ( !validWord( temp ) ) {
               cout << fatalString << messages[ 0 ] << termString << '\n';
               fatal = true;
            }
            else {
               *acPtr = temp;
               ++( *icPtr );
            }
            break;
         case BRANCH:
            *icPtr = *opPtr;
            break;
         case BRANCHNEG:
            *acPtr < 0 ? *icPtr = *opPtr : ++( *icPtr );
            break;
         case BRANCHZERO:
            *acPtr == 0 ? *icPtr = *opPtr : ++( *icPtr );
            break;
         case HALT:
            cout << "*** Simpletron execution terminated ***\n";
            break;
         default:
            cout << fatalString << messages[ 2 ] << termString << '\n';
            fatal = true;     
            break;
      }      
   } while ( *opCodePtr != HALT && !fatal );
   
   cout << "\n*************END SIMPLETRON EXECUTION*************\n";
}

void dump( const int * const memory, int accumulator, int instructionCounter,
          int instructionRegister, int operationCode, int operand )
{
   void output( const char * const, int, int, bool );   // prototype

   cout << "\nREGISTERS:\n";
   output( "accumulator", 5, accumulator, true );
   output( "instructionCounter", 2, instructionCounter, false );
   output( "instructionRegister", 5, instructionRegister, true );
   output( "operationCode", 2, operationCode, false );
   output( "operand", 2, operand, false );
   cout << "\n\nMEMORY:\n";

   int i = 0;
   cout << setfill( ' ' ) << setw( 3 ) << ' ';

   // print header
   for ( ; i <= 9; ++i )
      cout << setw( 5 ) << i << ' ';
      
   for ( i = 0; i < SIZE; ++i ) {
      if ( i % 10 == 0 )
         cout << '\n' << setw( 2 ) << i << ' ';

      cout << setiosflags( ios::internal | ios::showpos )
           << setw( 5 ) << setfill( '0' ) << memory[ i ] << ' ' 
           << resetiosflags( ios::internal | ios::showpos );
   }
   
   cout << endl;
}

bool validWord( int word )
{
   return word >= MIN_WORD && word <= MAX_WORD;
}

void output( const char * const sPtr, int width, int value, bool sign )
{        
   // format of "accumulator", etc.
   cout << setfill( ' ' ) << setiosflags( ios::left ) << setw( 20 )
        << sPtr << ' ';
   
   // is a +/- sign needed?
   if ( sign )
      cout << setiosflags( ios::showpos | ios::internal );

   // setup for displaying accumulator value, etc.
   cout << resetiosflags( ios::left ) << setfill( '0' );

   // determine the field widths and display value
   if ( width == 5 )
      cout << setw( width ) << value << '\n';
   else  // width is 2
      cout << setfill( ' ' ) << setw( 3 ) << ' ' << setw( width )
           << setfill( '0' ) << value << '\n';

   // disable sign if it was set
   if ( sign )
      cout << resetiosflags( ios::showpos | ios::internal );
}

/* Program OUTPUT :
***           Harish Welcome to Simpletron          ***
***            Loading File prog.sml                ***

00) 1007 (Read A)
01) 1008 (Read B)
02) 2007 (Load A)
03) 3008 (Add B)
04) 2109 (Store C)
05) 1109 (Write C)
06) 4300 (Halt)
07) 0 (Variable A)
08) 0 (Variable B)
09) 0 (Result C)
10) -99999 (Exit)

************START SIMPLETRON EXECUTION************

Enter an integer: 23
Enter an integer: 15
Contents of 09: 38
*** Simpletron execution terminated ***

*************END SIMPLETRON EXECUTION*************

REGISTERS:
accumulator          +0038
instructionCounter      06
instructionRegister  +4300
operationCode           43
operand                 00

MEMORY:
       0     1     2     3     4     5     6     7     8     9
 0 +1007 +1008 +2007 +3008 +2109 +1109 +4300 +0023 +0015 +0038
10 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000
20 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000
30 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000
40 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000
50 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000
60 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000
70 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000
80 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000
90 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000 +0000

*/

/* Input File: prog.sml
+1007 (Read A)
+1008 (Read B)
+2007 (Load A)
+3008 (Add B)
+2109 (Store C)
+1109 (Write C)
+4300 (Halt)
+0000 (Variable A)
+0000 (Variable B)
+0000 (Result C)
*/

